@@ -91,6 +91,7 @@ UINT32 c_vformat_length(const char *f, va_list args)
     {
       if (*f == '%')
 	{
+	  int do_hex = 0;
 	  while(*++f)
 	    {
 	      switch (*f)
@@ -99,6 +100,7 @@ UINT32 c_vformat_length(const char *f, va_list args)
 		  /* Do nothing */
 		  break;
 		case 'x':
+		  do_hex = 1;
 		  break;
 		default:
 		  goto end_options;
@@ -124,7 +126,7 @@ end_options:
 	      {
 	        UINT32 n = va_arg(args, UINT32);
 	        f++;
-                length += format_size_in_decimal(n);
+                length += (do_hex ? format_size_in_hex(n) : format_size_in_decimal(n));
   	        break;
 	      }
 	    case 's':
@@ -141,7 +143,7 @@ end_options:
 	    case 'S':
 	      {
 		struct ol_string *s = va_arg(args, struct ol_string *);
-		length += s->length;
+		length += s ? s->length : 6;
 		f++;
 		
 		break;
@@ -158,7 +160,8 @@ end_options:
               }
 	    case 'z':
 	      {
-		unsigned l = strlen(va_arg(args, char*));
+	        char *p = va_arg(args, char *);
+		unsigned l = p ? strlen(p) : 6;
 		length += l;
 
 		f++;
@@ -232,17 +235,31 @@ end_options:
 
 	    case 'i':
 	      {
+	      	UINT32 length;
 		UINT32 n = va_arg(args, UINT32);
-		
-                UINT32 length = format_size_in_decimal(n);
-	        unsigned i;
 
-		for (i = 0; i < length; i++)
-		  {
-		    buffer[length - i - 1] = '0' + n % 10;
-		    n /= 10;
-		  }
-		buffer += length;
+		if (!do_hex) {		
+		  unsigned i;
+
+		  length = format_size_in_decimal(n);
+		  for (i = 0; i < length; i++)
+		    {
+		      buffer[length - i - 1] = '0' + n % 10;
+		      n /= 10;
+		    }
+		}
+		else {
+		  char hexchars[] = "0123456789abcdef";
+		  unsigned i;
+		  
+		  length = format_size_in_hex(n);
+		  for (i = 0; i < length; i++)
+		    {
+		      buffer[length - i - 1] = hexchars[n & 0xf];
+		      n >>= 4;
+		    }
+		}
+  		buffer += length;
 		f++;
 		break;
 	      }
@@ -260,12 +277,19 @@ end_options:
 	    case 'S':
 	      {
 		struct ol_string *s = va_arg(args, struct ol_string *);
+		if (s) 
+		  {
+		    memcpy(buffer, s->data, s->length);
+		    buffer += s->length;
 
-		memcpy(buffer, s->data, s->length);
-		buffer += s->length;
-
-		if (do_free)
-		  ol_string_free(s);
+		    if (do_free)
+		      ol_string_free(s);
+		  }
+		else
+		  {
+		    memcpy(buffer, "(NULL)", 6);
+		    buffer += 6;
+		  }
 		f++;
 
 		break;
@@ -283,9 +307,12 @@ end_options:
 	    case 'z':
 	      {
 		char *s = va_arg(args, char *);
-		UINT32 length = strlen(s);
+		UINT32 length = s ? strlen(s) : 6;
 
-		memcpy(buffer, s, length);
+		if (s)
+		  memcpy(buffer, s, length);
+		else
+		  memcpy(buffer, "(NULL)", 6);
 		buffer += length;
 		f++;
 
@@ -341,6 +368,17 @@ UINT32 format_size_in_decimal(UINT32 n)
   return e+1;
 }
 
+UINT32 format_size_in_hex(UINT32 n)
+{
+  UINT32 mask = 0xf0000000;
+  UINT32 size = 8;
+  while (n & mask) {
+  	size--;
+  	mask >>= 4;
+  }
+  return size;
+}
+
 #if 0
 static int write_decimal_length(UINT8 *buffer, UINT32 n)
 {
