@@ -1,6 +1,6 @@
 /* io.c
  *
- * $Id: io.c,v 1.12 1999/05/29 13:57:30 bazsi Exp $ */
+ * $Id: io.c,v 1.13 1999/06/09 14:56:08 bazsi Exp $ */
 
 /* lsh, an implementation of the ssh protocol
  *
@@ -88,6 +88,7 @@ static void do_free_callouts(struct ol_queue *q)
 int io_iter(struct io_backend *b)
 {
 	unsigned long nfds; /* FIXME: Should be nfds_t if that type is defined */
+	unsigned long nco; /* number of callouts */
 	struct pollfd *fds = NULL;
 	int res = 0;
 	int timeout = -1;
@@ -171,6 +172,7 @@ closes
 				
 				fds[i].fd = fd->fd;
 				fds[i].events = 0;
+				fds[i].revents = 0;
 				
 				if (fd->want_read)
 					fds[i].events |= POLLIN;
@@ -203,7 +205,7 @@ closes
 			werror("io_iter: poll failed: %z", strerror(errno));
 			return 0;
 		}
-	if (nfds) {
+	if (nfds && res > 0) {
 	  /* Do io. Note that the callback functions may add new fds to the
 	   * head of the list, or clear the alive flag on any fd. */
 
@@ -213,10 +215,32 @@ closes
 		for(fd = b->files, i=0; fd; fd = fd->next, i++)
 		{
 			assert(i<nfds);
-			
+
+#if 0
+		        debug("io.c: poll for fd %i: events = 0x%xi, revents = 0x%xi.\n",
+                	      fds[i].fd, fds[i].events,fds[i].revents);
+#endif
+
 			if (!fd->super.alive)
 				continue;
 			
+			if (fds[i].revents & POLLHUP) {
+				if (fd->want_read && fd->read)
+					READ_FD(fd);
+				else if (fd->want_write && fd->write)
+					WRITE_FD(fd);
+				else {
+					werror("io_iter(): POLLHUP on inactive fd!\n");
+					closekill_fd(fd, CLOSE_EOF);
+				}
+				continue;
+			}
+			if (fds[i].revents & POLLPRI) {
+				closekill_fd(fd, CLOSE_PROTOCOL_FAILURE);
+				continue;
+			}
+			if (fds[i].revents & (POLLNVAL | POLLERR))
+				closekill_fd(fd, CLOSE_POLL_FAILED);
 			if (fds[i].revents & POLLOUT)
 				if (fd->want_write && fd->write)
 					WRITE_FD(fd);
@@ -224,17 +248,16 @@ closes
 			if (!fd->super.alive)
 				continue;
 			
-			if (fds[i].revents & (POLLIN | POLLHUP)) {
+			if (fds[i].revents & POLLIN) {
 				if (fd->want_read && fd->read)
 					READ_FD(fd);
 			}
 
-			if (fds[i].revents & (POLLNVAL | POLLERR))
-				closekill_fd(fd, CLOSE_POLL_FAILED);
 		}
 		assert(i == nfds);
 	}
 
+	nco = 0;
 	{
 		FOR_QUEUE(&b->callouts, struct callout *, n) {
 			if (now >= n->when) {
@@ -242,9 +265,15 @@ closes
 				ol_queue_remove(&n->header);
 				ol_space_free(n);
 			}
+			else {
+				nco++;
+			}
 		}
 	}
-	
+	if (!nfds && !nco) {
+		werror("io_iter(): no waiting fds and no callouts! exiting...\n");
+		return 0;
+	}
 	return 1;
 }
 
@@ -306,7 +335,7 @@ static int do_recv(struct abstract_read **r, UINT32 length, UINT8 *buffer, abstr
 	CAST(fd_read, closure, *r);
 
 	if (!length) {
-		werror("io.c: do_read(): Zero length read was requested.\n");
+		werror("io.c: do_recv(): Zero length read was requested.\n");
 		return 0;
 	}
 
@@ -449,8 +478,6 @@ static void listen_callback(struct nonblocking_fd *fd)
 	int res;
 	int conn;
 
-	/* FIXME: Do something with the peer address? */
-
 	conn = accept(fd->fd,
 		      (struct sockaddr *) &peer, &addr_len);
 	if (conn < 0)
